from scrapy.core.downloader.handlers.http10 import HTTP10DownloadHandler
from scrapy.core.downloader.handlers.http11 import (
    HTTP11DownloadHandler as HTTPDownloadHandler,
)

__all__ = [
    "HTTP10DownloadHandler",
    "HTTPDownloadHandler",
]
